/*
 * File: line.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

The functions in this file are a general set of line management utilities.
They are the only routines that touch the text. They also touch the buffer
and window structures, to make sure that the necessary updating gets done.
There are routines in this file that handle the kill buffer too. It isn't
here for any good reason.

Note that this code only updates the dot and mark values in the window list.
Since all the code acts on the current window, the buffer that we are
editing must be being displayed, which means that "b_nwnd" is non zero,
which means that the dot and mark values in the buffer headers are nonsense.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"line.h"
#include	"cursor.h"
#include	"killring.h"
#include	"messages.h"
#include	"utils.h"



#define	BSIZE(a)	(a + NBLOCK - 1) & (~(NBLOCK - 1))


/*++

This routine allocates a block of memory large enough to hold a LINE
containing "used" characters. Return a pointer to the new block, or
NULL if there isn't any memory left. Print a message in the message
line if no space.

++*/
LINE *alloc_line(used)
size_t used;

{
    LINE *lp;

    lp = (LINE *) malloc(sizeof(LINE) + used);

    lp->l_size = used;
    lp->l_used = used;
    return (lp);
}



/*++

Delete line "lp". Fix all of the links that might point at it (they are
moved to offset 0 of the next line. Unlink the line from whatever buffer it
might be in. Release the memory. The buffers are updated too.

++*/
void free_line(lp)
LINE *lp;
{
    BUFFER *bp;
    SCRN *scrp;                 /* screen to fix pointers in */
    WNDW *wp;
    int cmark;                  /* current mark */

    /* in all screens.... */
    scrp = first_screen;
    while (scrp) {
        wp = scrp->s_first_window;
        while (wp != NULL) {
            if (wp->w_linep == lp)
                wp->w_linep = lforw(lp);
            if (wp->w_dotp == lp) {
                wp->w_dotp = lforw(lp);
                wp->w_doto = 0;
            }
            for (cmark = 0; cmark < NMARKS; cmark++) {
                if (wp->w_markp[cmark] == lp) {
                    wp->w_markp[cmark] = lforw(lp);
                    wp->w_marko[cmark] = 0;
                }
            }
            wp = wp->w_wndp;
        }

        /* next screen! */
        scrp = scrp->s_next_screen;
    }

    bp = bheadp;
    while (bp != NULL) {
        if (bp->b_nwnd == 0) {
            if (bp->b_dotp == lp) {
                bp->b_dotp = lforw(lp);
                bp->b_doto = 0;
            }
            for (cmark = 0; cmark < NMARKS; cmark++) {
                if (bp->b_markp[cmark] == lp) {
                    bp->b_markp[cmark] = lforw(lp);
                    bp->b_marko[cmark] = 0;
                }
            }
        }
        bp = bp->b_bufp;
    }
    lp->l_bp->l_fp = lp->l_fp;
    lp->l_fp->l_bp = lp->l_bp;
    free((char *) lp);
    return;
}



/*++

This routine gets called when a character is changed in place in the
current buffer. It updates all of the required flags in the buffer and
window system. The flag used is passed as an argument; if the buffer is
being displayed in more than one window we change EDIT to HARD. Set MODE
if the mode line needs to be updated (the "*" for changed buffer has to
be set).

++*/
void change_line(flag)
int flag;
{
    WNDW *wp;
    SCRN *scrp;                 /* screen to fix pointers in */

    if (curbp->b_nwnd != 1)     /* Ensure hard.         */
        flag = WFHARD;
    if ((curbp->b_flag & BFCHG) == 0) { /* First change, so     */
        flag |= WFMODE;         /* update mode lines.   */
        curbp->b_flag |= BFCHG;
    }

    /* in all screens.... */
    scrp = first_screen;
    while (scrp) {
        /* make sure all the needed windows get this flag */
        wp = scrp->s_first_window;
        while (wp != NULL) {
            if (wp->w_bufp == curbp)
                wp->w_flag |= flag;
            wp = wp->w_wndp;
        }

        /* next screen! */
        scrp = scrp->s_next_screen;
    }
    return;
}


/*++

Insert a string at the current point.
Global helper and wrapped by insert_string.

++*/
int insertstring(instr)
char *instr;
{
    int status;

    status = TRUE;
    if (instr != NULL)
        while (*instr) {
            status =
                ((*instr == '\r') ? insertnewline() : insertnchars(1,
                                                                   *instr));

            /* Insertion error? */
            if (status != TRUE) {
                mlwrite(TEXT168);       /* "%%Can not insert string" */
                break;
            }
            instr++;
        }
    return (status);
}



/*++

Insert "n" copies of the character "c" at the current location of dot. In
the easy case all that happens is the text is stored in the line. In the
hard case, the line has to be reallocated. When the window list is updated,
take special care; I screwed it up once. You always update dot in the
current window. You update mark, and a dot in another window, if it is
greater than the place where you did the insert. Return TRUE if all is
well, and FALSE on errors.

++*/
int insertnchars(int n, char c)
{

    char *cp1;
    char *cp2;
    int cmark;                  /* current mark */
    int doto;
    int i;
    LINE *lp1;
    LINE *lp2;
    LINE *lp3;
    SCRN *scrp;                 /* screen to fix pointers in */
    WNDW *wp;

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    /* a zero insert means do nothing! */
    if (n == 0)
        return (TRUE);

    /* Negative numbers of inserted characters are right out! */
    if (n < 1)
        return (FALSE);

    /* mark the current window's buffer as changed */
    change_line(WFEDIT);

    lp1 = curwp->w_dotp;        /* Current line         */
    if (lp1 == curbp->b_linep) {        /* At the end: special  */
        if (curwp->w_doto != 0) {
            mlwrite(TEXT170);   /* "bug: insertnchars" */
            return (FALSE);
        }
        if ((lp2 = alloc_line(BSIZE(n))) == NULL)       /* Allocate new line    */
            return (FALSE);
        lp2->l_used = (size_t) n;
        lp3 = lp1->l_bp;        /* Previous line        */
        lp3->l_fp = lp2;        /* Link in              */
        lp2->l_fp = lp1;
        lp1->l_bp = lp2;
        lp2->l_bp = lp3;
        for (i = 0; i < n; ++i)
            lp2->l_text[i] = c;
        curwp->w_dotp = lp2;
        curwp->w_doto = n;
        return (TRUE);
    }
    doto = curwp->w_doto;       /* Save for later.      */
    if (lp1->l_used + n > lp1->l_size) {        /* Hard: reallocate     */
        if ((lp2 = alloc_line(BSIZE(lp1->l_used + n))) == NULL)
            return (FALSE);
        lp2->l_used = lp1->l_used + n;
        cp1 = &lp1->l_text[0];
        cp2 = &lp2->l_text[0];
        while (cp1 != &lp1->l_text[doto])
            *cp2++ = *cp1++;
        cp2 += n;
        while (cp1 != &lp1->l_text[lp1->l_used])
            *cp2++ = *cp1++;
        lp1->l_bp->l_fp = lp2;
        lp2->l_fp = lp1->l_fp;
        lp1->l_fp->l_bp = lp2;
        lp2->l_bp = lp1->l_bp;
        free((char *) lp1);
    } else {                    /* Easy: in place       */
        lp2 = lp1;              /* Pretend new line     */
        lp2->l_used += n;
        cp2 = &lp1->l_text[lp1->l_used];
        cp1 = cp2 - n;
        while (cp1 != &lp1->l_text[doto])
            *--cp2 = *--cp1;
    }
    for (i = 0; i < n; ++i)     /* Add the characters   */
        lp2->l_text[doto + i] = c;
    /* in all screens.... */
    scrp = first_screen;
    while (scrp) {

        wp = scrp->s_first_window;
        while (wp != NULL) {
            if (wp->w_linep == lp1)
                wp->w_linep = lp2;
            if (wp->w_dotp == lp1) {
                wp->w_dotp = lp2;
                if (wp == curwp || wp->w_doto > doto)
                    wp->w_doto += n;
            }
            for (cmark = 0; cmark < NMARKS; cmark++) {
                if (wp->w_markp[cmark] == lp1) {
                    wp->w_markp[cmark] = lp2;
                    if (wp->w_marko[cmark] > doto)
                        wp->w_marko[cmark] += n;
                }
            }
            wp = wp->w_wndp;
        }

        /* next screen! */
        scrp = scrp->s_next_screen;
    }
    return (TRUE);
}



/*++

Insert a newline into the buffer at the current location of dot in the
current window. The funny ass-backwards way it does things is not a botch;
it just makes the last line in the file not a special case. Return TRUE if
everything works out and FALSE on error (memory allocation failure). The
update of dot and mark is a bit easier then in the above case, because the
split forces more updating.
Global helper.

++*/
int insertnewline()
{

    char *cp1;
    char *cp2;
    int cmark;                  /* current mark */
    int doto;
    LINE *lp1;
    LINE *lp2;
    SCRN *scrp;                 /* screen to fix pointers in */
    WNDW *wp;

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    change_line(WFHARD);
    lp1 = curwp->w_dotp;        /* Get the address and  */
    doto = curwp->w_doto;       /* offset of "."        */
    if ((lp2 = alloc_line((size_t) doto)) == NULL)      /* New first half line  */
        return (FALSE);
    cp1 = &lp1->l_text[0];      /* Shuffle text around  */
    cp2 = &lp2->l_text[0];
    while (cp1 != &lp1->l_text[doto])
        *cp2++ = *cp1++;
    cp2 = &lp1->l_text[0];
    while (cp1 != &lp1->l_text[lp1->l_used])
        *cp2++ = *cp1++;
    lp1->l_used -= doto;
    lp2->l_bp = lp1->l_bp;
    lp1->l_bp = lp2;
    lp2->l_bp->l_fp = lp2;
    lp2->l_fp = lp1;

    /* in all screens.... */
    scrp = first_screen;
    while (scrp) {

        wp = scrp->s_first_window;
        while (wp != NULL) {
            if (wp->w_linep == lp1)
                wp->w_linep = lp2;
            if (wp->w_dotp == lp1) {
                if (wp->w_doto < doto)
                    wp->w_dotp = lp2;
                else
                    wp->w_doto -= doto;
            }
            for (cmark = 0; cmark < NMARKS; cmark++) {
                if (wp->w_markp[cmark] == lp1) {
                    if (wp->w_marko[cmark] < doto)
                        wp->w_markp[cmark] = lp2;
                    else
                        wp->w_marko[cmark] -= doto;
                }
            }
            wp = wp->w_wndp;
        }

        /* next screen! */
        scrp = scrp->s_next_screen;
    }
    return (TRUE);
}



/*++

Delete a newline. Join the current line with the next line. If the next line
is the magic header line always return TRUE; merging the last line with the
header line can be thought of as always being a successful operation, even
if nothing is done, and this makes the kill buffer work "right". Easy cases
can be done by shuffling data around. Hard cases require that lines be moved
about in memory. Return FALSE on error and TRUE if all looks ok.
Called only by line.deletenchars()

++*/
static int deletenewline()
{
    char *cp1;
    char *cp2;
    LINE *lp1;
    LINE *lp2;
    LINE *lp3;
    WNDW *wp;
    SCRN *scrp;                 /* screen to fix pointers in */
    int cmark;                  /* current mark */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    lp1 = curwp->w_dotp;
    lp2 = lp1->l_fp;
    if (lp2 == curbp->b_linep) {        /* At the buffer end.   */
        if (lp1->l_used == 0)   /* Blank line.          */
            free_line(lp1);
        return (TRUE);
    }
    if (lp2->l_used <= lp1->l_size - lp1->l_used) {
        cp1 = &lp1->l_text[lp1->l_used];
        cp2 = &lp2->l_text[0];
        while (cp2 != &lp2->l_text[lp2->l_used])
            *cp1++ = *cp2++;

        /* in all screens.... */
        scrp = first_screen;
        while (scrp) {

            wp = scrp->s_first_window;
            while (wp != NULL) {
                if (wp->w_linep == lp2)
                    wp->w_linep = lp1;
                if (wp->w_dotp == lp2) {
                    wp->w_dotp = lp1;
                    wp->w_doto += lp1->l_used;
                }
                for (cmark = 0; cmark < NMARKS; cmark++) {
                    if (wp->w_markp[cmark] == lp2) {
                        wp->w_markp[cmark] = lp1;
                        wp->w_marko[cmark] += lp1->l_used;
                    }
                }
                wp = wp->w_wndp;
            }

            /* next screen! */
            scrp = scrp->s_next_screen;
        }

        lp1->l_used += lp2->l_used;
        lp1->l_fp = lp2->l_fp;
        lp2->l_fp->l_bp = lp1;
        free((char *) lp2);
        return (TRUE);
    }
    if ((lp3 = alloc_line(lp1->l_used + lp2->l_used)) == NULL)
        return (FALSE);
    cp1 = &lp1->l_text[0];
    cp2 = &lp3->l_text[0];
    while (cp1 != &lp1->l_text[lp1->l_used])
        *cp2++ = *cp1++;
    cp1 = &lp2->l_text[0];
    while (cp1 != &lp2->l_text[lp2->l_used])
        *cp2++ = *cp1++;
    lp1->l_bp->l_fp = lp3;
    lp3->l_fp = lp2->l_fp;
    lp2->l_fp->l_bp = lp3;
    lp3->l_bp = lp1->l_bp;

    /* in all screens.... */
    scrp = first_screen;
    while (scrp) {

        wp = scrp->s_first_window;
        while (wp != NULL) {
            if (wp->w_linep == lp1 || wp->w_linep == lp2)
                wp->w_linep = lp3;
            if (wp->w_dotp == lp1)
                wp->w_dotp = lp3;
            else if (wp->w_dotp == lp2) {
                wp->w_dotp = lp3;
                wp->w_doto += lp1->l_used;
            }
            for (cmark = 0; cmark < NMARKS; cmark++) {
                if (wp->w_markp[cmark] == lp1)
                    wp->w_markp[cmark] = lp3;
                else if (wp->w_markp[cmark] == lp2) {
                    wp->w_markp[cmark] = lp3;
                    wp->w_marko[cmark] += lp1->l_used;
                }
            }
            wp = wp->w_wndp;
        }

        /* next screen! */
        scrp = scrp->s_next_screen;
    }

    free((char *) lp1);
    free((char *) lp2);
    return (TRUE);
}



/*++

This function deletes "n" bytes, starting at dot. Positive n
deletes forward, negative n deletes backwards. It understands how to
deal with end of lines, and with two byte characters. It returns TRUE
if all of the characters were deleted, and FALSE if they were not
(because dot ran into the buffer end). The "kflag" is TRUE if the text
should be put in the kill buffer.

++*/
int deletenchars(n, kflag)
long n;                         /* # of chars to delete */
int kflag;                      /* put killed text in kill buffer flag */

{
    char *cp1;
    char *cp2;
    LINE *dotp;
    int doto;
    int chunk;
    WNDW *wp;
    int cmark;                  /* current mark */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    /* going Forward? */
    if (n >= 0) {

        while (n > 0) {
            /* record the current point */
            dotp = curwp->w_dotp;
            doto = curwp->w_doto;

            /* can't delete past the end of the buffer */
            if (dotp == curbp->b_linep)
                return (FALSE);

            /* find out how many chars to delete on this line */
            chunk = (int) (dotp->l_used - doto);        /* Size of chunk.       */
            if (chunk > n)
                chunk = n;

            /* if at the end of a line, merge with the next */
            if (chunk == 0) {

                /* flag that we are making a hard change */
                change_line(WFHARD);
                if (deletenewline() == FALSE ||
                    (kflag != FALSE && kinsert(FORWARD, '\r') == FALSE))
                    return (FALSE);
                --n;
                continue;
            }

            /* flag the fact we are changing the current line */
            change_line(WFEDIT);

            /* find the limits of the kill */
            cp1 = &dotp->l_text[doto];
            cp2 = cp1 + chunk;

            /* save the text to the kill buffer */
            if (kflag != FALSE) {
                while (cp1 != cp2) {
                    if (kinsert(FORWARD, *cp1) == FALSE)
                        return (FALSE);
                    ++cp1;
                }
                cp1 = &dotp->l_text[doto];
            }

            /* copy what is left of the line upward */
            while (cp2 != &dotp->l_text[dotp->l_used])
                *cp1++ = *cp2++;
            dotp->l_used -= chunk;

            /* fix any other windows with the same text displayed */
            wp = wheadp;
            while (wp != NULL) {

                /* reset the dot if needed */
                if (wp->w_dotp == dotp && wp->w_doto >= doto) {
                    wp->w_doto -= chunk;
                    if (wp->w_doto < doto)
                        wp->w_doto = doto;
                }

                /* reset any marks if needed */
                for (cmark = 0; cmark < NMARKS; cmark++) {
                    if (wp->w_markp[cmark] == dotp
                        && wp->w_marko[cmark] >= doto) {
                        wp->w_marko[cmark] -= chunk;
                        if (wp->w_marko[cmark] < doto)
                            wp->w_marko[cmark] = doto;
                    }
                }

                /* onward to the next window */
                wp = wp->w_wndp;
            }

            /* indicate we have deleted chunk characters */
            n -= chunk;
        }
    } else {
        while (n < 0) {

            /* record the current point */
            dotp = curwp->w_dotp;
            doto = curwp->w_doto;

            /* can't delete past the beginning of the buffer */
            if (dotp == lforw(curbp->b_linep) && (doto == 0))
                return (FALSE);

            /* find out how many chars to delete on this line */
            chunk = doto;       /* Size of chunk.       */
            if (chunk > -n)
                chunk = -n;

            /* if at the beginning of a line, merge with the last */
            if (chunk == 0) {

                /* flag that we are making a hard change */
                change_line(WFHARD);
                previous_character(TRUE, 1);
                if (deletenewline() == FALSE ||
                    (kflag != FALSE && kinsert(REVERSE, '\r') == FALSE))
                    return (FALSE);
                ++n;
                continue;
            }

            /* flag the fact we are changing the current line */
            change_line(WFEDIT);

            /* find the limits of the kill */
            cp1 = &dotp->l_text[doto];
            cp2 = cp1 - chunk;

            /* save the text to the kill buffer */
            if (kflag != FALSE) {
                while (cp1 > cp2) {
                    if (kinsert(REVERSE, *(--cp1)) == FALSE)
                        return (FALSE);
                }
                cp1 = &dotp->l_text[doto];
            }

            /* copy what is left of the line downward */
            while (cp1 != &dotp->l_text[dotp->l_used])
                *cp2++ = *cp1++;
            dotp->l_used -= chunk;
            curwp->w_doto -= chunk;

            /* fix any other windows with the same text displayed */
            wp = wheadp;
            while (wp != NULL) {

                /* reset the dot if needed */
                if (wp->w_dotp == dotp && wp->w_doto >= doto) {
                    wp->w_doto -= chunk;
                    if (wp->w_doto < doto)
                        wp->w_doto = doto;
                }

                /* reset any marks if needed */
                for (cmark = 0; cmark < NMARKS; cmark++) {
                    if (wp->w_markp[cmark] == dotp
                        && wp->w_marko[cmark] >= doto) {
                        wp->w_marko[cmark] -= chunk;
                        if (wp->w_marko[cmark] < doto)
                            wp->w_marko[cmark] = doto;
                    }
                }

                /* onward to the next window */
                wp = wp->w_wndp;
            }

            /* indicate we have deleted chunk characters */
            n += chunk;
        }
    }
    return (TRUE);
}



/*++

Add a new line to the end of the indicated buffer.  Return FALSE if we
run out of memory Note that this works on non-displayed buffers as well.
Global helper.

++*/
int addline(BUFFER * bp, char *text)
{
    LINE *lp;
    size_t i;
    size_t ntext;

    /* allocate the memory to hold the line */
    ntext = strlen(text);
    if ((lp = alloc_line(ntext)) == NULL)
        return (FALSE);

    /* copy the text into the new line */
    for (i = 0; i < ntext; ++i)
        lputc(lp, i, text[i]);

    /* add the new line to the end of the buffer */
    bp->b_linep->l_bp->l_fp = lp;
    lp->l_bp = bp->b_linep->l_bp;
    bp->b_linep->l_bp = lp;
    lp->l_fp = bp->b_linep;

    /* if the point was at the end of the buffer,
       move it to the beginning of the new line */
    if (bp->b_dotp == bp->b_linep)
        bp->b_dotp = lp;
    return (TRUE);
}


/*++

Set fill column to n.

++*/
int set_fill_column(f, n)
int f, n;
{

    f = 0;                      /* splint */

    fillcol = n;
    mlwrite(TEXT59, n);         /* "[Fill column is %d]" */
    return (TRUE);
}



/*++

Get the a line number.

++*/
long getlinenum(bp, sline)
BUFFER *bp;                     /* buffer to get current line from */
LINE *sline;                    /* line to search for */
{
    LINE *lp;                   /* current line */
    long numlines;              /* # of lines before point */

    /* starting at the beginning of the buffer */
    lp = lforw(bp->b_linep);

    /* start counting lines */
    numlines = 0L;
    while (lp != bp->b_linep) {
        /* if we are on the current line, record it */
        if (lp == sline)
            break;
        ++numlines;
        lp = lforw(lp);
    }

    /* and return the resulting count */
    return (numlines + 1L);
}
